"""
Copyright(C) 2018-2021 Chris Hansen <hansec@uw.edu>
Copyright(C) 2021-2023 Stamus Networks
Written by Chris Hansen <hansec@uw.edu>
Written by Eric Leblond <eld@stamus-networks.com>

This file is part of Suricata Language Server.

Suricata Language Server is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Suricata Language Server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Suricata Language Server.  If not, see <http://www.gnu.org/licenses/>.
"""

import sys
import argparse
from suricatals.langserver import LangServer
from suricatals.jsonrpc import JSONRPC2Connection, ReadWriter
import json
__version__ = '0.9.5'


def error_exit(error_str):
    print("ERROR: {0}".format(error_str))
    sys.exit(-1)


def main():
    #
    parser = argparse.ArgumentParser()
    parser.description = "Suricata Language Server ({0})".format(__version__)
    parser.add_argument(
        '--version', action="store_true",
        help="Print server version number and exit"
    )
    parser.add_argument(
        '--suricata-binary', default="suricata",
        help="Path to Suricata binary"
    )
    parser.add_argument(
        '--suricata-config', default=None,
        help="Path to Suricata config"
    )
    parser.add_argument(
        '--debug-log', action="store_true",
        help="Generate debug log in project root folder"
    )
    parser.add_argument(
        '--max-lines', default=1000, type=int,
        help="Don't start suricata analysis over this file size"
    )
    parser.add_argument(
        '--max-tracked-files', default=100, type=int,
        help="Don't start suricata analysis if workspace file count is superior to this limit"
    )
    parser.add_argument(
        '--batch-file', default=None,
        help="Batch mode to parse only the file in argument"
    )
    parser.add_argument(
        '--no-engine-analysis', action='store_true', default=False,
        help='Disable suricata engine analysis (used with --batch-file only)'
    )
    args = parser.parse_args()
    if args.version:
        print("{0}".format(__version__))
        sys.exit(0)
    #
    settings = {
        "suricata_binary": args.suricata_binary,
        "suricata_config": args.suricata_config,
        "max_lines": args.max_lines,
        "max_tracked_files": args.max_tracked_files,
    }
    #
    if not args.batch_file and args.no_engine_analysis:
        print('--no-engine-analysis must be used with --batch-file')

    if args.batch_file is None:
        stdin, stdout = _binary_stdio()
        s = LangServer(conn=JSONRPC2Connection(ReadWriter(stdin, stdout)),
                       debug_log=args.debug_log, settings=settings)
        s.run()
    else:
        s = LangServer(conn=None, settings=settings)
        _, diags = s.analyse_file(args.batch_file, not args.no_engine_analysis)
        for diag in diags:
            print(json.dumps(diag.to_message()))


def _binary_stdio():
    """Construct binary stdio streams (not text mode).
    This seems to be different for Window/Unix Python2/3, so going by:
        https://stackoverflow.com/questions/2850893/reading-binary-data-from-stdin
    """

    stdin, stdout = sys.stdin.buffer, sys.stdout.buffer

    return stdin, stdout



"""
Copyright(C) 2018-2023 Stamus Networks
Written by Eric Leblond <el@stamus-networks.com>

This file is part of Suricata Language Server.

Suricata Language Server is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Suricata Language Server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Suricata Language Server.  If not, see <http://www.gnu.org/licenses/>.
"""

from json.decoder import JSONDecodeError
import subprocess
import tempfile
import shutil
import os
import json
import io
import re
import logging

log = logging.getLogger(__name__)


class TestRules():
    VARIABLE_ERROR = 101
    OPENING_RULE_FILE = 41  # Error when opening a file referenced in the source
    OPENING_DATASET_FILE = 322  # Error when opening a dataset referenced in the source
    USELESS_ERRNO = [40, 43, 44]
    CONFIG_FILE = """
%YAML 1.1
---
stats:
  enabled: no
logging:
  default-log-level: warning
  outputs:
  - console:
      enabled: yes
      type: json
app-layer:
  protocols:
    tls:
      ja3-fingerprints: yes
security:
  lua:
    allow-rules: yes
engine-analysis:
  rules-fast-pattern: yes
  rules: yes
vars:
  address-groups:
    HOME_NET: "[192.168.0.0/16,10.0.0.0/8,172.16.0.0/12]"
    EXTERNAL_NET: "!$HOME_NET"
    SCANNERS: "127.0.0.1"
    DC_SERVERS: "127.0.0.1"
    HTTP_SERVERS: "$HOME_NET"
    SMTP_SERVERS: "$HOME_NET"
    SQL_SERVERS: "$HOME_NET"
    DNS_SERVERS: "$HOME_NET"
    TELNET_SERVERS: "$HOME_NET"
    AIM_SERVERS: "$EXTERNAL_NET"
    DNP3_SERVER: "$HOME_NET"
    DNP3_CLIENT: "$HOME_NET"
    MODBUS_CLIENT: "$HOME_NET"
    MODBUS_SERVER: "$HOME_NET"
    ENIP_CLIENT: "$HOME_NET"
    ENIP_SERVER: "$HOME_NET"
  port-groups:
    HTTP_PORTS: "80"
    SHELLCODE_PORTS: "!80"
    ORACLE_PORTS: 1521
    SSH_PORTS: 22
    DNP3_PORTS: 20000
    MODBUS_PORTS: 502
    FILE_DATA_PORTS: "[$HTTP_PORTS,110,143]"
    FTP_PORTS: 21
    VXLAN_PORTS: 4789
    TEREDO_PORTS: 3544
"""

    REFERENCE_CONFIG = """
# config reference: system URL

config reference: bugtraq   http://www.securityfocus.com/bid/
config reference: bid	    http://www.securityfocus.com/bid/
config reference: cve       http://cve.mitre.org/cgi-bin/cvename.cgi?name=
#config reference: cve       http://cvedetails.com/cve/
config reference: secunia   http://www.secunia.com/advisories/

#whitehats is unfortunately gone
config reference: arachNIDS http://www.whitehats.com/info/IDS

config reference: McAfee    http://vil.nai.com/vil/content/v_
config reference: nessus    http://cgi.nessus.org/plugins/dump.php3?id=
config reference: url       http://
config reference: et        http://doc.emergingthreats.net/
config reference: etpro     http://doc.emergingthreatspro.com/
config reference: telus     http://
config reference: osvdb     http://osvdb.org/show/osvdb/
config reference: threatexpert http://www.threatexpert.com/report.aspx?md5=
config reference: md5	    http://www.threatexpert.com/report.aspx?md5=
config reference: exploitdb http://www.exploit-db.com/exploits/
config reference: openpacket https://www.openpacket.org/capture/grab/
config reference: securitytracker http://securitytracker.com/id?
config reference: secunia   http://secunia.com/advisories/
config reference: xforce    http://xforce.iss.net/xforce/xfdb/
config reference: msft      http://technet.microsoft.com/security/bulletin/
"""

    CLASSIFICATION_CONFIG = """
config classification: not-suspicious,Not Suspicious Traffic,3
config classification: unknown,Unknown Traffic,3
config classification: bad-unknown,Potentially Bad Traffic, 2
config classification: attempted-recon,Attempted Information Leak,2
config classification: successful-recon-limited,Information Leak,2
config classification: successful-recon-largescale,Large Scale Information Leak,2
config classification: attempted-dos,Attempted Denial of Service,2
config classification: successful-dos,Denial of Service,2
config classification: attempted-user,Attempted User Privilege Gain,1
config classification: unsuccessful-user,Unsuccessful User Privilege Gain,1
config classification: successful-user,Successful User Privilege Gain,1
config classification: attempted-admin,Attempted Administrator Privilege Gain,1
config classification: successful-admin,Successful Administrator Privilege Gain,1


# NEW CLASSIFICATIONS
config classification: rpc-portmap-decode,Decode of an RPC Query,2
config classification: shellcode-detect,Executable code was detected,1
config classification: string-detect,A suspicious string was detected,3
config classification: suspicious-filename-detect,A suspicious filename was detected,2
config classification: suspicious-login,An attempted login using a suspicious username was detected,2
config classification: system-call-detect,A system call was detected,2
config classification: tcp-connection,A TCP connection was detected,4
config classification: trojan-activity,A Network Trojan was detected, 1
config classification: unusual-client-port-connection,A client was using an unusual port,2
config classification: network-scan,Detection of a Network Scan,3
config classification: denial-of-service,Detection of a Denial of Service Attack,2
config classification: non-standard-protocol,Detection of a non-standard protocol or event,2
config classification: protocol-command-decode,Generic Protocol Command Decode,3
config classification: web-application-activity,access to a potentially vulnerable web application,2
config classification: web-application-attack,Web Application Attack,1
config classification: misc-activity,Misc activity,3
config classification: misc-attack,Misc Attack,2
config classification: icmp-event,Generic ICMP event,3
config classification: kickass-porn,SCORE! Get the lotion!,1
config classification: policy-violation,Potential Corporate Privacy Violation,1
config classification: default-login-attempt,Attempt to login by a default username and password,2

config classification: targeted-activity,Targeted Malicious Activity was Detected,1
config classification: exploit-kit,Exploit Kit Activity Detected,1
config classification: external-ip-check,Device Retrieving External IP Address Detected,2
config classification: domain-c2,Domain Observed Used for C2 Detected,1
config classification: pup-activity,Possibly Unwanted Program Detected,2
config classification: credential-theft,Successful Credential Theft Detected,1
config classification: social-engineering,Possible Social Engineering Attempted,2
config classification: coin-mining,Crypto Currency Mining Activity Detected,2
config classification: command-and-control,Malware Command and Control Activity Detected,1
"""
    SURICATA_SYNTAX_CHECK = "Suricata Syntax Check"
    SURICATA_ENGINE_ANALYSIS = "Suricata Engine Analysis"

    def __init__(self, suricata_binary='suricata', suricata_config=None) -> None:
        self.suricata_binary = suricata_binary
        self.suricata_config = suricata_config
        self.suricata_version = self.get_suricata_version()

    def get_suricata_version(self):
        suri_cmd = [self.suricata_binary, '-V']
        suriprocess = subprocess.Popen(suri_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        (outdata, _) = suriprocess.communicate()
        for line in io.StringIO(outdata.decode('utf-8')):
            mm = re.match(r'This is Suricata version (\d+\.\d+\.\d+)', line)
            if mm is not None:
                return mm.group(1)
        return "6.0.0"

    def json_compat_version(self):
        (major, minor, fix) = self.suricata_version.split('.')
        if int(major) < 6:
            return True
        elif int(major) == 6 and int(minor) == 0 and int(fix) < 4:
            return False
        return True

    def parse_suricata_error_after_7(self, error):
        ret = {
            'errors': [],
            'warnings': [],
        }
        error_stream = io.StringIO(error)
        wait_line = False
        prev_err = {}
        for line in error_stream:
            try:
                s_err = json.loads(line)
            except JSONDecodeError:
                continue
            if s_err['event_type'] != 'engine':
                continue
            s_err['engine']['source'] = self.SURICATA_SYNTAX_CHECK

            if not s_err['engine']['module'].startswith('detect') and s_err['engine']['module'] not in ['rule-vars']:
                continue

            if  re.search('Variable "(.+)" is not defined in configuration file', s_err['engine'].get('message', '')):
                s_err['engine']['variable_error'] = True

            if s_err['engine']['module'] == 'detect-parse':
                if s_err['log_level'] == 'Error':
                    ret['errors'].append(s_err['engine'])
                    wait_line = True
                else:
                    ret['warnings'].append(s_err['engine'])
                    getsid = re.compile(r"sid *:(\d+)")
                    match = getsid.search(line)
                    if match:
                        s_err['engine']['sid'] = int(match.groups()[0])
            elif s_err['engine']['module'] == 'detect-dataset':
                if not wait_line:
                    if s_err['engine']['message'].startswith("bad type"):
                        s_err['engine']['message'] = "dataset: " + s_err['engine']['message']
                    ret['errors'].append(s_err['engine'])
                    wait_line = True
            elif s_err['engine']['module'] == 'rules-vars':
                ret['errors'].append(s_err['engine'])
            elif s_err['engine']['module'] == 'detect':
                if 'error parsing signature' in s_err['engine']['message']:
                    message = s_err['engine']['message']
                    s_err['engine']['message'] = s_err['engine']['message'].split(' from file')[0]
                    getsid = re.compile(r"sid *:(\d+)")
                    match = getsid.search(line)
                    if match:
                        s_err['engine']['sid'] = int(match.groups()[0])
                    getline = re.compile(r"at line (\d+)$")
                    match = getline.search(message)
                    if match:
                        line_nb = int(match.groups()[0])
                        if wait_line:
                            if prev_err!= {}:
                                prev_err['engine']['line'] = line_nb - 1
                                if prev_err['engine'] not in ret['errors']:
                                    ret['errors'].append(prev_err['engine'])
                        else:
                            if prev_err != {} and prev_err['log_level'] == 'Warning':
                                prev_err['engine']['line'] = line_nb - 1
                                ret['errors'].append(prev_err['engine'])
                            else:
                                s_err['engine']['line'] = line_nb - 1
                                ret['errors'].append(s_err['engine'])
                    wait_line = False
                else:
                    ret['errors'].append(s_err['engine'])
            else:
                if not wait_line:
                    if s_err['log_level'] == 'Error':
                        ret['errors'].append(s_err['engine'])
                        wait_line = True
            prev_err = s_err
        return ret

    # pylint: disable=W0613
    def parse_suricata_error_before_7(self, error):
        ret = {
            'errors': [],
            'warnings': [],
        }
        files_list = []
        ignore_next = False
        error_stream = io.StringIO(error)
        error_type = 'errors'
        prev_err = None
        for line in error_stream:
            try:
                s_err = json.loads(line)
            except JSONDecodeError:
                continue
            s_err['engine']['source'] = self.SURICATA_SYNTAX_CHECK
            errno = s_err['engine']['error_code']
            if s_err.get('log_level', '') != 'Error':
                if errno not in [176, 242, 308]:
                    prev_err = s_err['engine']
                    continue
            if errno == self.VARIABLE_ERROR:
                s_err['engine']['suricata_error'] = True
                s_err['engine']['variable_error'] = True

                # suricata config is set when we should have all variables defined
                if self.suricata_config is None:
                    error_type = 'warnings'
                    s_err['engine']['warning'] = s_err['engine'].pop('error', '')
                ret[error_type].append(s_err['engine'])
                continue
            elif errno == self.OPENING_DATASET_FILE:
                m = re.match('fopen \'([^:]*)\' failed: No such file or directory', s_err['engine']['message'])
                if m is not None:
                    datasource = m.group(1)
                    s_err['engine']['message'] = 'Dataset source "%s" is a dependency " \
                        "and needs to be added to rulesets' % datasource
                    s_err['engine']['suricata_error'] = True
                    error_type = 'warnings'
                    ret[error_type].append(s_err['engine'])
                    ignore_next = True
                    continue
            elif errno == self.OPENING_RULE_FILE:
                m = re.match('opening hash file ([^:]*): No such file or directory', s_err['engine']['message'])
                if m is not None:
                    filename = m.group(1)
                    filename = filename.rsplit('/', 1)[1]
                    files_list.append(filename)
                    s_err['engine']['message'] = 'External file "%s" is a dependency ' \
                        'and needs to be added to rulesets' % filename
                    s_err['engine']['suricata_error'] = True
                    error_type = 'warnings'
                    ret[error_type].append(s_err['engine'])
                    continue
            elif errno == 176:
                warning, sig_content = s_err['engine']['message'].split('"', 1)
                ret['warnings'].append({'message': warning.rstrip(),
                                           'source': self.SURICATA_SYNTAX_CHECK,
                                           'content': sig_content.rstrip('"')})
            # Message for invalid signature
            elif errno == 276:
                rule, warning = s_err['engine']['message'].split(': ', 1)
                rule = int(rule.split(' ')[1])
                ret['warnings'].append({'message': warning.rstrip(), 'source': self.SURICATA_SYNTAX_CHECK, 'sid': rule})
            elif errno not in self.USELESS_ERRNO:
                # clean error message
                if errno == 39:
                    if 'failed to set up dataset' in s_err['engine']['message']:
                        if ignore_next:
                            continue
                    if ignore_next:
                        ignore_next = False
                        continue
                    if 'error parsing signature' in s_err['engine']['message']:
                        message = s_err['engine']['message']
                        s_err['engine']['message'] = s_err['engine']['message'].split(' from file')[0]
                        getsid = re.compile(r"sid *:(\d+)")
                        match = getsid.search(line)
                        if match:
                            s_err['engine']['sid'] = int(match.groups()[0])
                        getline = re.compile(r"at line (\d+)$")
                        match = getline.search(message)
                        if match:
                            line_nb = int(match.groups()[0])
                            if prev_err is not None:
                                prev_err['line'] = line_nb - 1
                                ret['errors'].append(prev_err)
                                prev_err = None
                            else:
                                if len(ret[error_type]):
                                    ret[error_type][-1]['line'] = line_nb - 1
                                error_type = 'errors'
                            continue
                if errno == 42:
                    s_err['engine']['message'] = s_err['engine']['message'].split(' from')[0]
                ret['errors'].append(s_err['engine'])
        return ret

    def parse_suricata_error(self, error):
        (major, _, _) = self.suricata_version.split('.')
        if int(major) < 7:
            return self.parse_suricata_error_before_7(error)
        else:
            return self.parse_suricata_error_after_7(error)

    def generate_config(self, tmpdir, config_buffer=None, related_files=None,
                        reference_config=None, classification_config=None, extra_conf=None):
        if not reference_config:
            reference_config = self.REFERENCE_CONFIG
        reference_file = os.path.join(tmpdir, "reference.config")
        with open(reference_file, 'w', encoding='utf-8') as rf:
            rf.write(reference_config)

        if not classification_config:
            classification_config = self.CLASSIFICATION_CONFIG
        classification_file = os.path.join(tmpdir, "classification.config")
        with open(classification_file, 'w', encoding='utf-8') as cf:
            cf.write(classification_config)

        if not config_buffer:
            if self.suricata_config is None:
                config_buffer = self.CONFIG_FILE
            else:
                with open(self.suricata_config, 'r', encoding='utf-8') as conf_file:
                    config_buffer = conf_file.read()
        config_file = os.path.join(tmpdir, "suricata.yaml")
        with open(config_file, 'w', encoding='utf-8') as cf:
            # write the config file in temp dir
            cf.write(config_buffer)
            cf.write("mpm-algo: ac-bs\n")
            cf.write("default-rule-path: " + tmpdir + "\n")
            cf.write("reference-config-file: " + tmpdir + "/reference.config\n")
            cf.write("classification-file: " + tmpdir + "/classification.config\n")
            if extra_conf:
                cf.write(extra_conf.format(tmpdir=tmpdir))
            cf.write("""
engine-analysis:
  rules-fast-pattern: yes
  rules: yes
logging:
  default-log-level: warning
  outputs:
  - console:
      enabled: yes
      type: json
stats:
  enabled: no
outputs:
  - eve-log:
    enabled: no
""")

        related_files = related_files or {}
        for rfile in related_files:
            related_file = os.path.join(tmpdir, rfile)
            with open(related_file, 'w', encoding='utf-8') as rf:
                rf.write(related_files[rfile])

        return config_file

    def _prepare_conf(self, rule_buffer, tmpdir, **kwargs):
        # write the rule file in temp dir
        rule_file = os.path.join(tmpdir, "file.rules")
        with open(rule_file, 'w', encoding='utf-8') as rf:
            rf.write(rule_buffer)

        if kwargs.get('extra_buffers'):
            for filename, content in kwargs['extra_buffers'].items():
                full_path = os.path.join(tmpdir, filename)
                with open(full_path, 'w', encoding="utf-8") as f:
                    f.write(content)

        return self.generate_config(
            tmpdir,
            config_buffer=kwargs.get('config_buffer'),
            related_files=kwargs.get('related_files'),
            reference_config=kwargs.get('reference_config'),
            classification_config=kwargs.get('classification_config'),
            extra_conf=kwargs.get('extra_conf')
        )

    def rules_infos(self, rule_buffer, **kwargs):
        tmpdir = tempfile.mkdtemp()
        config_file = self._prepare_conf(rule_buffer, tmpdir, **kwargs)
        rule_file = os.path.join(tmpdir, "file.rules")

        suri_cmd = [self.suricata_binary, '--engine-analysis', '-l', tmpdir, '-S', rule_file, '-c', config_file]
        suriprocess = subprocess.Popen(suri_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        suriprocess.communicate()

        res = {}
        json_path = os.path.join(tmpdir, 'rules.json')
        if os.path.exists(json_path):
            with open(json_path, 'r', encoding='utf-8') as f:
                for line in f.readlines():
                    content = json.loads(line)
                    res[content['id']] = content

        shutil.rmtree(tmpdir)
        return res

    def rule_buffer(self, rule_buffer, engine_analysis=True, config_buffer=None, related_files=None,
                    reference_config=None, classification_config=None, extra_buffers=None, **kwargs):
        tmpdir = tempfile.mkdtemp()
        config_file = self._prepare_conf(
            rule_buffer,
            tmpdir,
            config_buffer=config_buffer,
            related_files=related_files,
            reference_config=reference_config,
            classification_config=classification_config,
            extra_buffers=extra_buffers,
            **kwargs
        )

        rule_file = os.path.join(tmpdir, "file.rules")

        suri_cmd = [self.suricata_binary, '-T', '-l', tmpdir, '-S', rule_file, '-c', config_file]
        # start suricata in test mode
        suriprocess = subprocess.Popen(suri_cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        (errdata, _) = suriprocess.communicate()
        result = {'status': True, 'errors': "", 'warnings': [], 'info': []}
        # if not a success
        if suriprocess.returncode != 0:
            result['status'] = False
        result['errors'] = errdata.decode('utf-8')

        if engine_analysis:
            # runs rules analysis to have warnings
            suri_cmd = [self.suricata_binary, '--engine-analysis', '-l', tmpdir, '-S', rule_file, '-c', config_file]
            # start suricata in test mode
            suriprocess = subprocess.Popen(suri_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            suriprocess.communicate()
            engine_analysis = self.parse_engine_analysis(tmpdir)
            for signature in engine_analysis:
                for warning in signature.get('warnings', []):
                    result['warnings'].append({
                        'message': warning,
                        'source': self.SURICATA_ENGINE_ANALYSIS,
                        'sid': signature.get('sid', 'UNKNOWN'),
                        'content': signature['content']
                    })

                for info in signature.get('info', []):
                    result['info'].append({
                        'message': info,
                        'source': self.SURICATA_ENGINE_ANALYSIS,
                        'content': signature['content'],
                        'sid': signature.get('sid', 'UNKNOWN')
                    })

            mpm_analysis = self.mpm_parse_rules_json(tmpdir)
            result['mpm'] = mpm_analysis
        shutil.rmtree(tmpdir)
        return result

    def check_rule_buffer(self, rule_buffer, engine_analysis=True, config_buffer=None, related_files=None, extra_buffers=None, **kwargs):
        related_files = related_files or {}
        prov_result = self.rule_buffer(
            rule_buffer,
            engine_analysis=engine_analysis,
            config_buffer=config_buffer,
            related_files=related_files,
            extra_buffers=extra_buffers,
            **kwargs
        )

        if len(prov_result.get('errors', '')):
            res = self.parse_suricata_error(prov_result['errors'])
            if 'errors' in res:
                prov_result['errors'] = res['errors']
            if 'warnings' in res:
                prov_result['warnings'].extend(res['warnings'])
        return prov_result

    def parse_engine_analysis(self, log_dir):
        if self.json_compat_version():
            json_path = os.path.join(log_dir, 'rules.json')
            if os.path.isfile(json_path):
                return self.parse_engine_analysis_v2(json_path)
            else:
                # we end up in this case when no rules is valid in buffer
                return []
        return self.parse_engine_analysis_v1(log_dir)

    def parse_engine_analysis_v1(self, log_dir):
        analysis = []
        analysis_path = os.path.join(log_dir, 'rules_analysis.txt')
        if not os.path.isfile(analysis_path):
            return analysis
        with open(analysis_path, 'r', encoding='utf-8') as analysis_file:
            in_sid_data = False
            signature = {}
            for line in analysis_file:
                if line.startswith("== "):
                    in_sid_data = True
                    signature = {'sid': line.split(' ')[2]}
                    continue
                elif in_sid_data and len(line) == 1:
                    in_sid_data = False
                    analysis.append(signature)
                    signature = {}
                elif in_sid_data and 'content' not in signature:
                    signature['content'] = line.strip()
                    continue
                elif in_sid_data and 'Warning: ' in line:
                    warning = line.split('Warning: ')[1]
                    if 'warnings' not in signature:
                        signature['warnings'] = []
                    signature['warnings'].append(warning.strip())
                elif in_sid_data and 'Fast Pattern' in line:
                    if 'info' not in signature:
                        signature['info'] = []
                    signature['info'].append(line.strip())
        return analysis

    def parse_engine_analysis_v2(self, json_path):
        analysis = []
        with open(json_path, 'r', encoding='utf-8') as analysis_file:
            for line in analysis_file:
                signature_info = {}
                try:
                    signature_info = json.loads(line)
                except JSONDecodeError:
                    pass
                signature_msg = {'content': signature_info['raw']}
                if 'type' in signature_info:
                    if 'info' not in signature_msg:
                        signature_msg['info'] = []
                    type_msg = f'Rule type is "{signature_info["type"]}"'
                    signature_msg['info'].append(type_msg)
                if 'id' in signature_info:
                    signature_msg['sid'] = signature_info['id']
                if 'flags' in signature_info:
                    if 'toserver' in signature_info['flags'] and 'toclient' in signature_info['flags']:
                        if 'warnings' not in signature_msg:
                            signature_msg['warnings'] = []
                        signature_msg['warnings'].append('Rule inspect server and client side, consider adding a flow keyword')
                if 'warnings' in signature_info:
                    if 'warnings' not in signature_msg:
                        signature_msg['warnings'] = []
                    signature_msg['warnings'].extend(signature_info.get('warnings', []))
                if 'notes' in signature_info:
                    if 'info' not in signature_msg:
                        signature_msg['info'] = []
                    signature_msg['info'].extend(signature_info.get('notes', []))
                if 'engines' in signature_info:
                    app_proto = None
                    multiple_app_proto = False
                    got_raw_match = False
                    got_content = False
                    got_pcre = False
                    for engine in signature_info['engines']:
                        if 'app_proto' in engine:
                            if app_proto is None:
                                app_proto = engine.get('app_proto')
                            else:
                                if app_proto != engine.get('app_proto'):
                                    if app_proto not in ['http', 'http2', 'dns', 'doh2'] or engine.get('app_proto') not in ['http', 'http2', 'dns', 'doh2']:
                                        multiple_app_proto = True

                        else:
                            got_raw_match = True
                        for match in engine.get('matches', []):
                            if match['name'] == 'content':
                                got_content = True
                            elif match['name'] == 'pcre':
                                got_pcre = True
                    if got_pcre and not got_content:
                        if 'warnings' not in signature_msg:
                            signature_msg['warnings'] = []
                        signature_msg['warnings'].append('Rule with pcre without content match (possible performance issue)')
                    if app_proto is not None and got_raw_match:
                        if 'warnings' not in signature_msg:
                            signature_msg['warnings'] = []
                        signature_msg['warnings'].append('Application layer "%s" combined with raw match, '
                                                         'consider using a match on application buffer' % (app_proto))
                    if multiple_app_proto:
                        if 'warnings' not in signature_msg:
                            signature_msg['warnings'] = []
                        signature_msg['warnings'].append('Multiple application layers in same signature')
                analysis.append(signature_msg)
        return analysis

    def mpm_parse_rules_json(self, log_dir):
        mpm_data = []
        mpm_analysis = {'buffer': {}, 'sids': {}}
        try:
            with open(os.path.join(log_dir, 'rules.json'), 'r', encoding='utf-8') as rules_json:
                for line in rules_json:
                    # some suricata version have an invalid JSON formatted message
                    try:
                        rule_analysis = json.loads(line)
                    except json.JSONDecodeError:
                        return None
                    if 'mpm' in rule_analysis:
                        rule_analysis['mpm']['id'] = rule_analysis['id']
                        rule_analysis['mpm']['gid'] = rule_analysis['gid']
                        mpm_data.append(rule_analysis['mpm'])
                    else:
                        if 'engines' in rule_analysis:
                            fp_buffer = None
                            fp_pattern = None
                            for engine in rule_analysis.get('engines', []):
                                if engine['is_mpm']:
                                    fp_buffer = engine['name']
                                    for match in engine.get('matches', []):
                                        if match.get('name') == 'content':
                                            if match.get('content', {}).get('is_mpm', False):
                                                fp_pattern = match['content']['pattern']
                                                break
                                    if fp_pattern:
                                        break
                            if fp_buffer and fp_pattern:
                                mpm_data.append({'id': rule_analysis['id'], 'gid': rule_analysis['gid'],
                                                 'buffer': fp_buffer, 'pattern': fp_pattern})
                            continue
                        if 'lists' in rule_analysis:
                            fp_buffer = None
                            fp_pattern = None
                            for key in rule_analysis['lists']:
                                fp_buffer = key
                                for match in rule_analysis['lists'][key].get('matches', []):
                                    if match.get('name') == 'content':
                                        if match.get('content', {}).get('is_mpm', False):
                                            fp_pattern = match['content']['pattern']
                                            break
                                if fp_pattern:
                                    break
                            if fp_buffer and fp_pattern:
                                mpm_data.append({'id': rule_analysis['id'], 'gid': rule_analysis['gid'],
                                                 'buffer': fp_buffer, 'pattern': fp_pattern})
                            continue
        except FileNotFoundError:
            return mpm_analysis
        # target to have
        # { 'http.host': { 'grosminet': { 'count': 34, sigs: [{'id': 2, 'gid':1}]} } }
        for sig in mpm_data:
            if sig['buffer'] in mpm_analysis['buffer']:
                if sig['pattern'] in mpm_analysis['buffer'][sig['buffer']]:
                    mpm_analysis['buffer'][sig['buffer']][sig['pattern']]['count'] += 1
                    mpm_analysis['buffer'][sig['buffer']][sig['pattern']]['sigs'].append({'id': sig['id'], 'gid': sig['gid']})
                else:
                    mpm_analysis['buffer'][sig['buffer']][sig['pattern']] = {'count': 1,
                                                                             'sigs': [{'id': sig['id'], 'gid': sig['gid']}]}
            else:
                mpm_analysis['buffer'][sig['buffer']] = {sig['pattern']: {'count': 1,
                                                                          'sigs': [{'id': sig['id'], 'gid': sig['gid']}]}}
            mpm_analysis['sids'][sig['id']] = {'buffer': sig['buffer'], 'pattern': sig['pattern']}
        return mpm_analysis

    def build_keywords_list(self):
        tmpdir = tempfile.mkdtemp()
        config_file = self.generate_config(tmpdir)
        suri_cmd = [self.suricata_binary, '--list-keywords=csv', '-l', tmpdir, '-c', config_file]
        # start suricata in test mode
        suriprocess = subprocess.Popen(suri_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        (outdata, _) = suriprocess.communicate()
        shutil.rmtree(tmpdir)
        keywords = outdata.decode('utf-8').splitlines()
        keywords.pop(0)
        keywords_list = []
        for keyword in keywords:
            keyword_array = keyword.split(';')
            try:
                detail = 'No option'
                if 'sticky' in keyword_array[3]:
                    detail = 'Sticky Buffer'
                elif keyword_array[3] == 'none':
                    detail = 'No option'
                else:
                    detail = keyword_array[3]
                documentation = keyword_array[1]
                if len(keyword_array) > 5:
                    if 'https' in keyword_array[4]:
                        documentation += "\n\n"
                        documentation += "[Documentation](" + keyword_array[4] + ")"
                        documentation = {'kind': 'markdown', 'value': documentation}
                keyword_item = {'label': keyword_array[0], 'kind': 14, 'detail': detail, 'documentation': documentation}
                if 'content modifier' in keyword_array[3]:
                    keyword_item['tags'] = [1]
                    keyword_item['detail'] = 'Content Modifier'
                keywords_list.append(keyword_item)
            except IndexError:
                pass
        return keywords_list

    def build_app_layer_list(self):
        tmpdir = tempfile.mkdtemp()
        config_file = self.generate_config(tmpdir)
        suri_cmd = [self.suricata_binary, '--list-app-layer-proto', '-l', tmpdir, '-c', config_file]
        # start suricata in test mode
        suriprocess = subprocess.Popen(suri_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        (outdata, _) = suriprocess.communicate()
        shutil.rmtree(tmpdir)
        applayers = outdata.decode('utf-8').splitlines()
        while not applayers[0].startswith("===="):
            applayers.pop(0)
        applayers.pop(0)
        applayers_list = [{'label': 'tcp', 'detail': 'tcp', 'kind': 14}, {'label': 'udp', 'detail': 'udp', 'kind': 14}]
        for app_layer in applayers:
            app_layer_item = {'label': app_layer, 'detail': app_layer, 'kind': 14}
            applayers_list.append(app_layer_item)
        return applayers_list



"""
Copyright(C) 2021-2023 Stamus Networks
Written by Eric Leblond <el@stamus-networks.com>

This file is part of Suricata Language Server.

Suricata Language Server is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Suricata Language Server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Suricata Language Server.  If not, see <http://www.gnu.org/licenses/>.
"""

import hashlib

import re

from .lsp_helpers import Diagnosis, FileRange


class Signature:
    GETSID = re.compile(r"sid *:(\d+)")
    GETREV = re.compile(r"rev *:(\d+)")
    SIG_END = re.compile(r"\) *$")
    SIG_CONTENT = re.compile(r"content *:")

    def __init__(self, line, content, multiline=False):
        self.line = line
        self.line_end = line
        self.multiline = multiline
        if self.multiline:
            self.content = content.strip('\\')
        else:
            self.content = content
        self.raw_content = [content]
        self.multiline = multiline
        self.sid = 0
        self._search_sid(content)
        self.rev = 0
        self._search_rev(content)
        self.mpm = None
        self.has_error = False

    def append_content(self, content, line):
        self.content += content.rstrip('\\')
        self.raw_content.append(content)
        self.line_end = line
        if self.sid == 0:
            self._search_sid(content)
        if self.rev == 0:
            self._search_rev(content)

    def _get_diag_range_by_sid(self):
        fr = None
        i = 0
        for line in self.raw_content:
            if "sid:" in line:
                line_start = self.line + i
                line_end = line_start
                range_start = line.index('sid:')
                range_end = range_start + len('sid:')
                fr = FileRange(line_start, range_start, line_end, range_end)
                break
            i += 1
        return fr

    def get_diag_range(self, mode="all", pattern=""):
        fr = None
        if mode == "all":
            last_char = len(self.raw_content[-1].rstrip())
            fr = FileRange(self.line, 0, self.line_end, last_char)
        elif mode == "sid":
            fr = self._get_diag_range_by_sid()
        elif mode == "pattern":
            pattern_match = re.compile(f'content: *("{re.escape(pattern)}")')
            i = 0
            found = False
            for line in self.raw_content:
                match = pattern_match.search(line)
                if match:
                    line_start = self.line + i
                    line_end = line_start
                    range_start = match.start()
                    range_end = match.end()
                    fr = FileRange(line_start, range_start, line_end, range_end)
                    found = True
                    break
                i += 1
            if found is False:
                fr = self._get_diag_range_by_sid()
        return fr

    def get_content_keyword_count(self):
        count = 0
        for line in self.raw_content:
            count += len(self.SIG_CONTENT.findall(line))
        return count

    def sls_syntax_check(self):
        diagnosis = []
        # check for incomplete signature
        if self.SIG_END.search(self.raw_content[-1]) is None:
            sig_range = self.get_diag_range(mode="all")
            if sig_range:
                end_diag = Diagnosis()
                end_diag.range = sig_range
                end_diag.message = "Missing closing parenthesis: incomplete signature"
                end_diag.severity = Diagnosis.WARNING_LEVEL
                end_diag.source = "SLS syntax check"
                end_diag.sid = self.sid,
                end_diag.content = self.content
                diagnosis.append(end_diag)
        return diagnosis

    def _search_sid(self, content):
        match = self.GETSID.search(content)
        if match:
            self.sid = int(match.groups()[0])

    def _search_rev(self, content):
        match = self.GETREV.search(content)
        if match:
            self.rev = int(match.groups()[0])

    def __repr__(self):
        return "Signature()"

    def __str__(self):
        return "%d:%s" % (self.sid, self.content)


class SignatureSet:
    def __init__(self):
        self.content_map = {}
        self.line_map = {}
        self.sid_map = {}
        self.signatures = []

    def add_signature(self, line, content, multiline=False):
        signature = Signature(line, content, multiline=multiline)
        self.signatures.append(signature)
        self.content_map[content] = signature
        self.line_map[line] = signature
        if signature.sid:
            if signature.sid in self.sid_map:
                if self.sid_map[signature.sid].rev > signature.rev:
                    signature.has_error = True
                elif self.sid_map[signature.sid].rev < signature.rev:
                    self.sid_map[signature.sid].has_error = True
                self.sid_map[signature.sid] = signature
            else:
                self.sid_map[signature.sid] = signature
        return signature

    def add_content_to_signature(self, sig_line, line, content):
        signature = self.get_sig_by_line(sig_line)
        if signature is None:
            return
        signature.append_content(content, line)
        self.content_map[signature.content] = signature
        if signature.sid != 0:
            self.sid_map[signature.sid] = signature

    def get_sig_by_line(self, line):
        return self.line_map.get(line)

    def get_sig_by_content(self, content):
        return self.content_map.get(content)

    def get_sig_by_sid(self, sid):
        return self.sid_map.get(sid)


class SuricataFile:
    IS_COMMENT = re.compile(r"[ \t]*#")
    GET_MULTILINES = re.compile(r"\\ *$")

    def __init__(self, path, rules_tester, empty=False):
        self.path = path
        self.rules_tester = rules_tester
        self.contents_split = []
        self.sigset = SignatureSet()
        self.nLines = 0
        self.hash = None
        self.mpm = {}
        self.diagnosis = None
        if not empty:
            self.count_lines()

    def count_lines(self):
        with open(self.path, 'r', encoding='utf-8', errors='replace') as fhandle:
            content = fhandle.read()
            self.nLines = len(content.splitlines())

    def copy(self):
        """Copy content to new file object (does not copy objects)"""
        copy_obj = SuricataFile(self.path, self.rules_tester)
        return copy_obj

    def load_from_disk(self):
        """Read file from disk"""
        try:
            contents = ''
            with open(self.path, 'r', encoding='utf-8', errors='replace') as fhandle:
                contents = fhandle.read()
            self.hash = hashlib.md5(contents.encode('utf-8')).hexdigest()
            self.contents_split = contents.splitlines()
            self.nLines = len(self.contents_split)
            self.parse_file()
        # pylint: disable=W0703
        except Exception:
            return 'Could not read/decode file'
        else:
            return None

    def sort_diagnosis(self, key):
        return -key.severity

    def check_file(self, workspace=None, engine_analysis=True, **kwargs):
        diagnostics = []
        result = {}
        if not workspace:
            workspace = {}
        with open(self.path, 'r', encoding='utf-8', errors='replace') as fhandle:
            result = self.rules_tester.check_rule_buffer(fhandle.read(), engine_analysis, **kwargs)
            self.mpm = result.get('mpm', {}).get('buffer')
            for sid in result.get('mpm', {}).get('sids', []):
                signature = self.sigset.get_sig_by_sid(sid)
                if signature is not None:
                    signature.mpm = result.get('mpm', {}).get('sids', {}).get(sid)
        for error in result.get('errors', []):
            if 'line' in error:
                l_diag = Diagnosis()
                l_diag.message = error['message']
                l_diag.source = error['source']
                l_diag.severity = Diagnosis.ERROR_LEVEL
                signature = self.sigset.get_sig_by_line(error['line'])
                if signature:
                    signature.has_error = True
                    sig_range = signature.get_diag_range(mode="all")
                    l_diag.range = sig_range
                    l_diag.content = signature.content
                    l_diag.sid = signature.sid
                else:
                    e_range = FileRange(error['line'], 0, error['line'], 1000)
                    l_diag.range = e_range
                    l_diag.content = error.get('content', '')
                    l_diag.sid = error.get('sid', 'UNKNOWN')

                diagnostics.append(l_diag)
        for warning in result.get('warnings', []):
            line = None
            signature = None
            l_diag = Diagnosis()
            l_diag.message = warning['message']
            l_diag.source = warning['source']
            l_diag.severity = Diagnosis.WARNING_LEVEL
            if 'line' in warning:
                line = warning['line']
                signature = self.sigset.get_sig_by_line(line)
            elif 'content' in warning:
                signature = self.sigset.get_sig_by_content(warning['content'])
                if signature is not None:
                    line = signature.line
            elif 'sid' in warning:
                signature = self.sigset.get_sig_by_sid(warning['sid'])
                if signature is not None:
                    line = signature.line
            if line is None:
                continue
            if signature is not None:
                l_diag.range = signature.get_diag_range(mode="sid")
                if warning.get('suricata_error', False):
                    signature.has_error = True

                l_diag.content = signature.content
                l_diag.sid = signature.sid
            else:
                w_range = FileRange(line, 0, line, 1000)
                l_diag.range = w_range
                l_diag.content = warning.get('content', '')
                l_diag.sid = warning.get('sid', 'UNKNOWN')

            diagnostics.append(l_diag)

        if engine_analysis:
            for info in result.get('info', []):
                line = None
                signature = None
                l_diag = Diagnosis()
                l_diag.message = info['message']
                l_diag.source = info['source']
                l_diag.severity = Diagnosis.INFO_LEVEL
                if 'line' in info:
                    line = info['line']
                elif 'content' in info:
                    signature = self.sigset.get_sig_by_content(info['content'])
                    if signature:
                        line = signature.line
                if line is None:
                    continue
                sig_range = FileRange(line, 0, line, 1)
                if signature is not None:
                    if "Fast Pattern \"" in info['message']:
                        if signature.mpm is not None:
                            sig_range = signature.get_diag_range(mode='pattern', pattern=signature.mpm['pattern'])
                        else:
                            sig_range = signature.get_diag_range(mode='sid')
                    else:
                        sig_range = signature.get_diag_range(mode='sid')

                    l_diag.content = signature.content
                    l_diag.sid = signature.sid
                else:
                    l_diag.content = info.get('content', '')
                    l_diag.sid = info.get('sid', 'UNKNOWN')

                l_diag.range = sig_range
                diagnostics.append(l_diag)
            for sig in self.sigset.signatures:
                if sig.mpm is None:
                    if sig.sid and sig.has_error is False:
                        message = "No Fast Pattern used, if possible add one content match to improve performance."
                        l_diag = Diagnosis()
                        l_diag.message = message
                        l_diag.source = "Suricata MPM Analysis"
                        l_diag.severity = Diagnosis.INFO_LEVEL
                        l_diag.range = sig.get_diag_range(mode="sid")
                        l_diag.sid = sig.sid
                        l_diag.content = sig.content
                        diagnostics.append(l_diag)
                    continue
                # mpm is content:"$pattern"
                pattern = self.mpm.get(sig.mpm['buffer'], {}).get(sig.mpm['pattern'])
                if pattern is None:
                    continue
                pattern_count = pattern['count']
                for sig_file in workspace:
                    if sig_file != self.path:
                        file_obj = workspace.get(sig_file)
                        if file_obj is None or file_obj.mpm is None:
                            continue
                        f_pattern = file_obj.mpm.get(sig.mpm['buffer'], {}).get(sig.mpm['pattern'])
                        if f_pattern is None:
                            continue
                        pattern_count += f_pattern['count']
                l_diag = Diagnosis()
                if pattern_count > 1:
                    l_diag.message = "Fast Pattern '%s' on '%s' buffer is used in %d different signatures, " \
                                     "consider using a unique fast pattern to improve performance." \
                                     % (sig.mpm['pattern'], sig.mpm['buffer'], pattern_count)
                    l_diag.source = "SLS MPM Analysis"
                else:
                    if sig.get_content_keyword_count() == 1:
                        continue
                    l_diag.message = "Fast Pattern '%s' on '%s' buffer" % (sig.mpm['pattern'], sig.mpm['buffer'])
                    l_diag.source = "Suricata MPM Analysis"
                sig_range = sig.get_diag_range(mode="pattern", pattern=sig.mpm['pattern'])
                l_diag.severity = Diagnosis.INFO_LEVEL
                l_diag.range = sig.get_diag_range(mode="pattern", pattern=sig.mpm['pattern'])
                l_diag.sid = sig.sid
                l_diag.content = sig.content
                diagnostics.append(l_diag)
        for sig in self.sigset.signatures:
            sls_diag = sig.sls_syntax_check()
            if len(sls_diag):
                diagnostics.extend(sls_diag)
        self.diagnosis = diagnostics
        return result['status'], sorted(diagnostics, key=self.sort_diagnosis)

    def parse_file(self):
        """Build file Info by parsing file"""
        i = 0
        self.sigset = SignatureSet()
        multi_lines_index = -1
        for content_line in self.contents_split:
            if self.IS_COMMENT.match(content_line):
                i += 1
                continue
            if multi_lines_index >= 0:
                self.sigset.add_content_to_signature(multi_lines_index, i, content_line)
                if self.GET_MULTILINES.search(content_line):
                    i += 1
                    continue
                else:
                    multi_lines_index = -1
                    i += 1
                    continue
            elif self.GET_MULTILINES.search(content_line):
                multi_lines_index = i
                self.sigset.add_signature(i, content_line, multiline=True)
                i += 1
                continue
            else:
                if len(content_line) > 0 and not content_line.isspace():
                    self.sigset.add_signature(i, content_line, multiline=False)
            i += 1

    def apply_change(self, content_update):
        self.contents_split = content_update['text'].splitlines()
        self.nLines = len(self.contents_split)
        self.parse_file()



"""
Copyright(C) 2018-2021 Chris Hansen <hansec@uw.edu>
Copyright(C) 2021-2022 Stamus Networks

This file is part of Suricata Language Server.

Suricata Language Server is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Suricata Language Server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Suricata Language Server.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
import os
import traceback
import re

from suricatals.jsonrpc import path_from_uri
from suricatals.parse_signatures import SuricataFile
from suricatals.tests_rules import TestRules

log = logging.getLogger(__name__)

SURICATA_RULES_EXT_REGEX = re.compile(r'^\.rules?$', re.I)


def init_file(filepath, rules_tester, line_limit):
    file_obj = SuricataFile(filepath, rules_tester)
    if file_obj.nLines < line_limit:
        file_obj.check_file()
    return file_obj, None


class LangServer:
    ACTIONS_ITEMS = [
        {'label': 'alert', 'kind': 14, 'detail': 'Alert action', 'documentation': 'Trigger alert'},
        {'label': 'config', 'kind': 14, 'detail': 'Alert action',
         'documentation': 'Configuration signature. Used mostly for conditional logging.'},
        {'label': 'drop', 'kind': 14, 'detail': 'Alert action', 'documentation': 'Trigger alert and drop flow'},
        {'label': 'pass', 'kind': 14, 'detail': 'Alert action', 'documentation': 'Stop inspecting the data'},
        {'label': 'reject', 'kind': 14, 'detail': 'Alert action', 'documentation': 'Trigger alert and reset session'},
        {'label': 'rejectsrc', 'kind': 14, 'detail': 'Alert action',
            'documentation': 'Trigger alert and reset session for source IP'},
        {'label': 'rejectdst', 'kind': 14, 'detail': 'Alert action',
            'documentation': 'Trigger alert and reset session for destination IP'},
        {'label': 'rejectboth', 'kind': 14, 'detail': 'Alert action',
            'documentation': 'Trigger alert and reset session for both IPs'},
    ]

    def __init__(self, conn, debug_log=False, settings=None):
        self.conn = conn
        self.running = True
        self.root_path = None
        self.fs = None
        self.workspace = {}
        self.source_dirs = []
        self.excl_paths = []
        self.excl_suffixes = []
        self.post_messages = []
        self.streaming = True
        self.debug_log = debug_log
        # Get launch settings
        if settings is None:
            settings = {}
        self.nthreads = settings.get("nthreads", 4)
        self.notify_init = settings.get("notify_init", False)
        self.sync_type = settings.get("sync_type", 1)
        self.suricata_binary = settings.get("suricata_binary", 'suricata')
        self.suricata_config = settings.get("suricata_config", None)
        self.max_lines = settings.get("max_lines", 1000)
        self.max_tracked_files = settings.get("max_tracked_files", 100)
        self.rules_tester = TestRules(suricata_binary=self.suricata_binary, suricata_config=self.suricata_config)
        self.keywords_list = self.rules_tester.build_keywords_list()
        self.app_layer_list = self.rules_tester.build_app_layer_list()

    def post_message(self, message, msg_type=1):
        self.conn.send_notification("window/showMessage", {
            "type": msg_type,
            "message": message
        })

    def run(self):
        # Run server
        while self.running:
            try:
                request = self.conn.read_message()
                self.handle(request)
            except EOFError:
                break
            # pylint: disable=W0703
            except Exception as e:
                log.error("Unexpected error: %s", e, exc_info=True)
                break
            else:
                for message in self.post_messages:
                    self.post_message(message[1], message[0])
                self.post_messages = []

    def handle(self, request):
        def noop(_):
            return None
        # Request handler
        log.debug("REQUEST %s %s", request.get("id"), request.get("method"))
        handler = {
            "initialize": self.serve_initialize,
            "textDocument/documentSymbol": noop,
            "textDocument/completion": self.serve_autocomplete,
            "textDocument/signatureHelp": noop,
            "textDocument/definition": noop,
            "textDocument/references": noop,
            "textDocument/hover": noop,
            "textDocument/implementation": noop,
            "textDocument/rename": noop,
            "textDocument/didOpen": self.serve_onOpen,
            "textDocument/didSave": self.serve_onSave,
            "textDocument/didClose": self.serve_onClose,
            "textDocument/didChange": self.serve_onChange,
            "textDocument/codeAction": noop,
            "initialized": noop,
            "workspace/didChangeWatchedFiles": noop,
            "workspace/symbol": noop,
            "$/cancelRequest": noop,
            "shutdown": noop,
            "exit": self.serve_exit,
        }.get(request["method"], self.serve_default)
        # handler = {
        #     "workspace/symbol": self.serve_symbols,
        # }.get(request["method"], self.serve_default)
        # We handle notifications differently since we can't respond
        if "id" not in request:
            try:
                handler(request)
            # pylint: disable=W0703
            except Exception:
                log.warning(
                    "error handling notification %s", request, exc_info=True)
            return
        #
        try:
            resp = handler(request)
        except JSONRPC2Error as e:
            self.conn.write_error(
                request["id"], code=e.code, message=e.message, data=e.data)
            log.warning("RPC error handling request %s", request, exc_info=True)
        # pylint: disable=W0703
        except Exception as e:
            self.conn.write_error(
                request["id"],
                code=-32603,
                message=str(e),
                data={
                    "traceback": traceback.format_exc(),
                })
            log.warning("error handling request %s", request, exc_info=True)
        else:
            self.conn.write_response(request["id"], resp)

    def serve_initialize(self, request):
        # Setup language server
        params = request["params"]
        self.root_path = path_from_uri(
            params.get("rootUri") or params.get("rootPath") or "")
        self.source_dirs.append(self.root_path)
        # Recursively add sub-directories
        if len(self.source_dirs) == 1:
            self.source_dirs = []
            for dirName, subdirList, fileList in os.walk(self.root_path):
                if self.excl_paths.count(dirName) > 0:
                    while (len(subdirList) > 0):
                        del subdirList[0]
                    continue
                contains_source = False
                for filename in fileList:
                    _, ext = os.path.splitext(os.path.basename(filename))
                    if SURICATA_RULES_EXT_REGEX.match(ext):
                        contains_source = True
                        break
                if contains_source:
                    self.source_dirs.append(dirName)
        # Initialize workspace
        self.workspace_init()
        #
        server_capabilities = {
            "completionProvider": {
                "resolveProvider": False,
                "triggerCharacters": ["%"]
            },
            # "definitionProvider": True,
            # "documentSymbolProvider": True,
            # "referencesProvider": True,
            # "hoverProvider": True,
            # "implementationProvider": True,
            # "renameProvider": True,
            # "workspaceSymbolProvider": True,
            "textDocumentSync": self.sync_type
        }
        if self.notify_init:
            self.post_messages.append([3, "suricatals initialization complete"])
        return {"capabilities": server_capabilities}
        #     "workspaceSymbolProvider": True,
        #     "streaming": False,
        # }

    def _initial_params_autocomplete(self, request, file_obj):
        params = request["params"]
        edit_index = params['position']['line']
        sig_content = file_obj.contents_split[edit_index]
        sig_index = params['position']['character']
        word_split = re.split(' +', sig_content[0:sig_index])
        if len(word_split) == 1:
            return self.ACTIONS_ITEMS
        if len(word_split) == 2:
            return self.app_layer_list
        if edit_index == 0:
            return None
        elif not re.search(r'\\ *$', file_obj.contents_split[edit_index - 1]):
            return None

    def serve_autocomplete(self, request):
        params = request["params"]
        uri = params["textDocument"]["uri"]
        path = path_from_uri(uri)
        file_obj = self.workspace.get(path)
        if file_obj is None:
            return None
        edit_index = params['position']['line']
        sig_content = file_obj.contents_split[edit_index]
        sig_index = params['position']['character']
        log.debug(sig_content)
        # not yet in content matching so just return nothing
        if '(' not in sig_content[0:sig_index]:
            return self._initial_params_autocomplete(request, file_obj)
        cursor = sig_index - 1
        while cursor > 0:
            log.debug("At index: %d of %d (%s)", cursor, len(sig_content), sig_content[cursor:sig_index])
            if not sig_content[cursor].isalnum() and not sig_content[cursor] in ['.', '_']:
                break
            cursor -= 1
        log.debug("Final is: %d : %d", cursor, sig_index)
        if cursor == sig_index - 1:
            return None
        # this is an option edit so dont list keyword
        if sig_content[cursor] in [':', ',']:
            return None
        cursor += 1
        partial_keyword = sig_content[cursor:sig_index]
        log.debug("Got keyword start: '%s'", partial_keyword)
        items_list = []
        for item in self.keywords_list:
            if item['label'].startswith(partial_keyword):
                items_list.append(item)
        if len(items_list):
            return items_list
        return None

    def send_diagnostics(self, uri):
        diag_results, diag_exp = self.get_diagnostics(uri)
        if diag_results is not None:
            self.conn.send_notification("textDocument/publishDiagnostics", {
                "uri": uri,
                "diagnostics": diag_results
            })
        elif diag_exp is not None:
            self.conn.write_error(
                -1,
                code=-32603,
                message=str(diag_exp),
                data={
                    "traceback": traceback.format_exc(),
                })

    def get_diagnostics(self, uri):
        filepath = path_from_uri(uri)
        file_obj = self.workspace.get(filepath)
        if file_obj is not None and file_obj.nLines < self.max_lines:
            try:
                _, diags_list = file_obj.check_file(workspace=self.workspace)
                diags = [diag.to_message() for diag in diags_list]
            # pylint: disable=W0703
            except Exception as e:
                if os.path.isfile(file_obj.path):
                    return None, e
            else:
                return diags, None
        return None, None

    def serve_onChange(self, request):
        # Update workspace from file sent by editor
        params = request["params"]
        uri = params["textDocument"]["uri"]
        path = path_from_uri(uri)
        file_obj = self.workspace.get(path)
        if file_obj is None:
            self.post_message('Change request failed for unknown file "{0}"'.format(path))
            log.error('Change request failed for unknown file "%s"', path)
            return
        else:
            # Update file contents with changes
            reparse_req = True
            if self.sync_type == 1:
                file_obj.apply_change(params["contentChanges"][0])
            else:
                try:
                    reparse_req = False
                    for change in params["contentChanges"]:
                        reparse_flag = file_obj.apply_change(change)
                        reparse_req = (reparse_req or reparse_flag)
                # pylint: disable=W0703
                except Exception:
                    self.post_message('Change request failed for file "{0}": Could not apply change'.format(path))
                    log.error('Change request failed for file "%s": Could not apply change', path, exc_info=True)
                    return
        # Parse newly updated file
        if reparse_req:
            _, err_str = self.update_workspace_file(path)
            if err_str is not None:
                self.post_message('Change request failed for file "{0}": {1}'.format(path, err_str))

    def serve_onOpen(self, request):
        self.serve_onSave(request, did_open=True)

    def serve_onClose(self, request):
        self.serve_onSave(request, did_close=True)

    def serve_onSave(self, request, did_open=False, did_close=False):
        # Update workspace from file on disk
        params = request["params"]
        uri = params["textDocument"]["uri"]
        filepath = path_from_uri(uri)
        # Skip update and remove objects if file is deleted
        if did_close and (not os.path.isfile(filepath)):
            return
        did_change, err_str = self.update_workspace_file(filepath, read_file=True, allow_empty=did_open)
        if err_str is not None:
            self.post_message('Save request failed for file "{0}": {1}'.format(filepath, err_str))
            return
        if did_change:
            self.send_diagnostics(uri)

    def update_workspace_file(self, filepath, read_file=False, allow_empty=False):
        # Update workspace from file contents and path
        try:
            file_obj = self.workspace.get(filepath)
            if read_file:
                if file_obj is None:
                    # Create empty file if not yet saved to disk
                    if not os.path.isfile(filepath):
                        file_obj = SuricataFile(filepath, self.rules_tester, empty=True)
                        if allow_empty:
                            self.workspace[filepath] = file_obj
                            return False, None
                        else:
                            return False, 'File does not exist'  # Error during load
                    else:
                        file_obj = SuricataFile(filepath, self.rules_tester)
                hash_old = file_obj.hash
                err_string = None
                if os.path.isfile(filepath):
                    err_string = file_obj.load_from_disk()
                    file_obj.parse_file()
                if err_string is not None:
                    log.error("%s: %s", err_string, filepath)
                    return False, err_string  # Error during file read
                if hash_old == file_obj.hash:
                    return False, None
        # pylint: disable=W0703
        except Exception:
            log.error("Error while parsing file %s", filepath, exc_info=True)
            return False, 'Error during parsing'  # Error during parsing
        if filepath not in self.workspace:
            self.workspace[filepath] = file_obj
        return True, None

    def workspace_init(self):
        # Get filenames
        file_list = []
        for source_dir in self.source_dirs:
            for filename in os.listdir(source_dir):
                _, ext = os.path.splitext(os.path.basename(filename))
                if SURICATA_RULES_EXT_REGEX.match(ext):
                    filepath = os.path.normpath(os.path.join(source_dir, filename))
                    if self.excl_paths.count(filepath) > 0:
                        continue
                    inc_file = True
                    for excl_suffix in self.excl_suffixes:
                        if filepath.endswith(excl_suffix):
                            inc_file = False
                            break
                    if inc_file:
                        file_list.append(filepath)
        # Process files
        # don't send to analysis if too many files
        if len(file_list) > self.max_tracked_files:
            return
        from multiprocessing import Pool
        pool = Pool(processes=self.nthreads)
        results = {}
        for filepath in file_list:
            results[filepath] = pool.apply_async(init_file, args=(filepath, self.rules_tester, self.max_lines))
        pool.close()
        pool.join()
        for path, result in results.items():
            result_obj = result.get()
            if result_obj[0] is None:
                self.post_messages.append([1, 'Initialization failed for file "{0}": {1}'.format(path, result_obj[1])])
                continue
            self.workspace[path] = result_obj[0]

    def serve_exit(self, _):
        # Exit server
        self.workspace = {}
        self.running = False

    def serve_default(self, request):
        # Default handler (errors!)
        raise JSONRPC2Error(
            code=-32601,
            message="method {} not found".format(request["method"]))

    def analyse_file(self, filepath, engine_analysis=True, **kwargs):
        file_obj = SuricataFile(filepath, self.rules_tester)
        file_obj.load_from_disk()
        return file_obj.check_file(engine_analysis=engine_analysis, **kwargs)

    def rules_infos(self, rule_buffer, **kwargs):
        return self.rules_tester.rules_infos(rule_buffer, **kwargs)


class JSONRPC2Error(Exception):
    def __init__(self, code, message, data=None):
        super().__init__(message)
        self.code = code
        self.message = message
        self.data = data



import unittest
from suricatals.langserver import LangServer

class TestSyntax(unittest.TestCase):

    def test_cleanrules(self):
        s = LangServer(conn=None, settings=None)
        _, diags = s.analyse_file("tests/clean.rules")
        self.assertTrue(len(diags) == 3)
        for diag in diags:
            self.assertTrue(diag.severity == 4)

    def test_invalid_multilines(self):
        s = LangServer(conn=None, settings=None)
        _, diags = s.analyse_file("tests/invalid-multiline.rules")
        self.assertTrue(len(diags) == 2)
        has_missing = False
        for diag in diags:
            if diag.severity == 2 and diag.message.startswith("Missing closing"):
                has_missing = True
        self.assertTrue(has_missing)

    def test_missing_incomplete(self):
        s = LangServer(conn=None, settings=None)
        _, diags = s.analyse_file("tests/missing-incomplete.rules")
        has_missing = 0
        for diag in diags:
            if diag.severity == 2 and diag.message.startswith("Missing closing"):
                has_missing += 1
        self.assertEqual(has_missing, 2)

    def test_fast_pattern_analysis(self):
        s = LangServer(conn=None, settings=None)
        _, diags = s.analyse_file("tests/fast-pattern-analysis.rules")
        self.assertEqual(len(diags), 3)
        for diag in diags:
            self.assertEqual(diag.severity, 4)
            if diag.range.line_start in [0, 1]:
                self.assertTrue('is used in' in diag.message)
            if diag.range.line_start == 2:
                self.assertTrue(diag.message.startswith("Fast Pattern"))
                self.assertFalse('is used in' in diag.message)

    def test_invalid_http_host(self):
        s = LangServer(conn=None, settings=None)
        _, diags = s.analyse_file("tests/invalid-http-host.rules")
        self.assertEqual(len(diags), 1)
        for diag in diags:
            self.assertEqual(diag.severity, 1)

    def test_pattern_error(self):
        s = LangServer(conn=None, settings=None)
        _, diags = s.analyse_file("tests/pattern-syntax.rules")
        self.assertEqual(len(diags), 3)
        for diag in diags:
            self.assertEqual(diag.severity, 4)

    def test_sig_shadow(self):
        s = LangServer(conn=None, settings=None)
        _, diags = s.analyse_file("tests/sig-shadow.rules")
        self.assertEqual(len(diags), 1)
        for diag in diags:
            self.assertEqual(diag.severity, 2)


if __name__ == '__main__':
    unittest.main()
